# Copyright 2020-present Kensho Technologies, LLC.
# @generated
# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
import grpc

from google.protobuf import wrappers_pb2 as google_dot_protobuf_dot_wrappers__pb2


class TestServiceStub(object):
  """Simple service to test the interceptors
  """

  def __init__(self, channel):
    """Constructor.

    Args:
      channel: A grpc.Channel.
    """
    self.UnaryUnary = channel.unary_unary(
        '/eagr.TestService/UnaryUnary',
        request_serializer=google_dot_protobuf_dot_wrappers__pb2.StringValue.SerializeToString,
        response_deserializer=google_dot_protobuf_dot_wrappers__pb2.StringValue.FromString,
        )
    self.UnaryStream = channel.unary_stream(
        '/eagr.TestService/UnaryStream',
        request_serializer=google_dot_protobuf_dot_wrappers__pb2.StringValue.SerializeToString,
        response_deserializer=google_dot_protobuf_dot_wrappers__pb2.StringValue.FromString,
        )
    self.StreamUnary = channel.stream_unary(
        '/eagr.TestService/StreamUnary',
        request_serializer=google_dot_protobuf_dot_wrappers__pb2.StringValue.SerializeToString,
        response_deserializer=google_dot_protobuf_dot_wrappers__pb2.StringValue.FromString,
        )
    self.StreamStream = channel.stream_stream(
        '/eagr.TestService/StreamStream',
        request_serializer=google_dot_protobuf_dot_wrappers__pb2.StringValue.SerializeToString,
        response_deserializer=google_dot_protobuf_dot_wrappers__pb2.StringValue.FromString,
        )


class TestServiceServicer(object):
  """Simple service to test the interceptors
  """

  def UnaryUnary(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def UnaryStream(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def StreamUnary(self, request_iterator, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def StreamStream(self, request_iterator, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')


def add_TestServiceServicer_to_server(servicer, server):
  rpc_method_handlers = {
      'UnaryUnary': grpc.unary_unary_rpc_method_handler(
          servicer.UnaryUnary,
          request_deserializer=google_dot_protobuf_dot_wrappers__pb2.StringValue.FromString,
          response_serializer=google_dot_protobuf_dot_wrappers__pb2.StringValue.SerializeToString,
      ),
      'UnaryStream': grpc.unary_stream_rpc_method_handler(
          servicer.UnaryStream,
          request_deserializer=google_dot_protobuf_dot_wrappers__pb2.StringValue.FromString,
          response_serializer=google_dot_protobuf_dot_wrappers__pb2.StringValue.SerializeToString,
      ),
      'StreamUnary': grpc.stream_unary_rpc_method_handler(
          servicer.StreamUnary,
          request_deserializer=google_dot_protobuf_dot_wrappers__pb2.StringValue.FromString,
          response_serializer=google_dot_protobuf_dot_wrappers__pb2.StringValue.SerializeToString,
      ),
      'StreamStream': grpc.stream_stream_rpc_method_handler(
          servicer.StreamStream,
          request_deserializer=google_dot_protobuf_dot_wrappers__pb2.StringValue.FromString,
          response_serializer=google_dot_protobuf_dot_wrappers__pb2.StringValue.SerializeToString,
      ),
  }
  generic_handler = grpc.method_handlers_generic_handler(
      'eagr.TestService', rpc_method_handlers)
  server.add_generic_rpc_handlers((generic_handler,))



# Copyright 2020-present Kensho Technologies, LLC.
import unittest

from ...server.middleware import MetricsMiddleware


class TestMiddlewares(unittest.TestCase):
    def test_MetricsMiddleware_get_decorator_naming(self):
        metrics_middleware = MetricsMiddleware()

        # Assert that no exception is raised
        metrics_middleware.get_decorator("/in/need.of_normalization", {})

        with self.assertRaises(AssertionError):
            metrics_middleware.get_decorator("-no_dash_at_start", {})



# Copyright 2020-present Kensho Technologies, LLC.
from unittest.mock import MagicMock

from google.protobuf import descriptor_pb2
from grpc_reflection.v1alpha import reflection_pb2_grpc


SERVICE_A = MagicMock()
SERVICE_A.name = "service_a"
SERVICE_B = MagicMock()
SERVICE_B.name = "service_b"
SERVICES_MOCK = MagicMock()
SERVICES_MOCK.list_services_response = MagicMock()
SERVICES_MOCK.list_services_response.service = [SERVICE_A, SERVICE_B]

PROTO_A = descriptor_pb2.FileDescriptorProto(name="proto_a")
PROTO_BYTES_A = PROTO_A.SerializeToString()
PROTO_B = descriptor_pb2.FileDescriptorProto(name="proto_b", dependency=["proto_c", "proto_d"])
PROTO_BYTES_B = PROTO_B.SerializeToString()
PROTO_C = descriptor_pb2.FileDescriptorProto(name="proto_c")
PROTO_BYTES_C = PROTO_C.SerializeToString()
PROTO_D = descriptor_pb2.FileDescriptorProto(name="proto_d", dependency=["proto_e", "proto_f"])
PROTO_BYTES_D = PROTO_D.SerializeToString()
PROTO_E = descriptor_pb2.FileDescriptorProto(name="proto_e")
PROTO_BYTES_E = PROTO_E.SerializeToString()
PROTO_F = descriptor_pb2.FileDescriptorProto(name="proto_f")
PROTO_BYTES_F = PROTO_F.SerializeToString()

RESPONSE_MOCK_A = MagicMock()
RESPONSE_MOCK_A.file_descriptor_response = MagicMock()
RESPONSE_MOCK_A.file_descriptor_response.file_descriptor_proto = [PROTO_BYTES_A]
RESPONSE_MOCK_B = MagicMock()
RESPONSE_MOCK_B.file_descriptor_response = MagicMock()
RESPONSE_MOCK_B.file_descriptor_response.file_descriptor_proto = [PROTO_BYTES_B]
RESPONSE_MOCK_C = MagicMock()
RESPONSE_MOCK_C.file_descriptor_response = MagicMock()
RESPONSE_MOCK_C.file_descriptor_response.file_descriptor_proto = [PROTO_BYTES_C]
RESPONSE_MOCK_D = MagicMock()
RESPONSE_MOCK_D.file_descriptor_response = MagicMock()
RESPONSE_MOCK_D.file_descriptor_response.file_descriptor_proto = [PROTO_BYTES_D]
RESPONSE_MOCK_E = MagicMock()
RESPONSE_MOCK_E.file_descriptor_response = MagicMock()
RESPONSE_MOCK_E.file_descriptor_response.file_descriptor_proto = [PROTO_BYTES_E]
RESPONSE_MOCK_F = MagicMock()
RESPONSE_MOCK_F.file_descriptor_response = MagicMock()
RESPONSE_MOCK_F.file_descriptor_response.file_descriptor_proto = [PROTO_BYTES_F]


def get_file_by_filename(request):
    if request.file_by_filename == "proto_a":
        return RESPONSE_MOCK_A
    elif request.file_by_filename == "proto_b":
        return RESPONSE_MOCK_B
    elif request.file_by_filename == "proto_c":
        return RESPONSE_MOCK_C
    elif request.file_by_filename == "proto_d":
        return RESPONSE_MOCK_D
    elif request.file_by_filename == "proto_e":
        return RESPONSE_MOCK_E
    elif request.file_by_filename == "proto_f":
        return RESPONSE_MOCK_F
    else:
        raise AssertionError("Cannot find file request named {}".format(request.file_by_filename))


def get_file_containing_symbol(request):
    if request.file_containing_symbol == "service_a":
        return RESPONSE_MOCK_A
    if request.file_containing_symbol == "service_b":
        return RESPONSE_MOCK_B
    else:
        raise AssertionError(
            "Cannot find file request with symbol {}".format(request.file_containing_symbol)
        )


def server_reflection_info_mock(reflection_requests):
    responses = []
    for request in reflection_requests:
        if getattr(request, "file_by_filename", None):
            responses.append(get_file_by_filename(request))
        if getattr(request, "file_containing_symbol", None):
            responses.append(get_file_containing_symbol(request))
        if getattr(request, "list_services", None):
            if request.list_services == "services":
                responses.append(SERVICES_MOCK)
            else:
                raise AssertionError(
                    "Listing services for value {} not allowed".format(request.list_services)
                )

    return responses


reflection_client_mock = MagicMock(spec=reflection_pb2_grpc.ServerReflectionStub)
reflection_client_mock.ServerReflectionInfo = server_reflection_info_mock



# Copyright 2020-present Kensho Technologies, LLC.
"""Implementing client-side grpc interceptors"""
import functools
import json

import backoff
import grpc
import prometheus_client


CLIENTSIDE_METRICS_HISTO = prometheus_client.Histogram(
    "clientside_grpc_endpoint",
    "Response time histogram for grpc endpoints from the client-side",
    labelnames=("client_name", "server_name", "service", "endpoint"),
)
CLIENTSIDE_ERROR_COUNTER = prometheus_client.Counter(
    "clientside_grpc_endpoint_error",
    "Clientside exception counts for grpc methods",
    labelnames=("client_name", "server_name", "service", "endpoint", "exception"),
)

GRPC_RENDEZVOUS_ERROR = "_Rendezvous"


def get_service_and_method_from_url(method_url):
    """Extract service and method names from the method url string.

    Returns strings that are applicable as prometheus metrics and/or labels.

    Args:
        method_url: string

    Returns:
        tuple(service_name, method_name)
    """
    name_parts = method_url.split("/")
    if len(name_parts) != 3 or name_parts[0] != "" or name_parts[1] == "" or name_parts[2] == "":
        raise AssertionError("Invalid method name: {}".format(method_url))

    return (name_parts[1].replace(".", "_"), name_parts[2].replace(".", "_"))


class GRPCClientGeneralInterceptor(
    grpc.UnaryUnaryClientInterceptor,
    grpc.StreamUnaryClientInterceptor,
    grpc.UnaryStreamClientInterceptor,
    grpc.StreamStreamClientInterceptor,
):
    """General GRPC client interceptor that intercepts all functions."""

    def __init__(self, decorator_fn):
        """Initialize interceptor with a factory function producing decorators."""
        super(GRPCClientGeneralInterceptor, self).__init__()
        self._decorator_fn = decorator_fn

    def _intercept_call(self, continuation, client_call_details, request_or_iterator):
        """Interceptor implementation."""
        metadata = _get_metadata_map_from_client_details(client_call_details)
        decorator = self._decorator_fn(client_call_details.method, metadata)
        if not decorator:
            handler = continuation
        else:
            handler = decorator(continuation)

        return handler(client_call_details, request_or_iterator)

    def intercept_unary_unary(self, continuation, client_call_details, request):
        """Intercept unary-unary."""
        return self._intercept_call(continuation, client_call_details, request)

    def intercept_stream_unary(self, continuation, client_call_details, request_iterator):
        """Intercept stream-unary."""
        return self._intercept_call(continuation, client_call_details, request_iterator)

    def intercept_unary_stream(self, continuation, client_call_details, request):
        """Intercept unary-stream."""
        return self._intercept_call(continuation, client_call_details, request)

    def intercept_stream_stream(self, continuation, client_call_details, request_iterator):
        """Intercept stream-stream."""
        return self._intercept_call(continuation, client_call_details, request_iterator)


class GRPCClientUnaryOutputInterceptor(
    grpc.UnaryUnaryClientInterceptor, grpc.StreamUnaryClientInterceptor
):
    """GRPC interceptor that makes intercepts only unary-output grpcs."""

    def __init__(self, decorator_fn):
        """Initialize interceptor with a factory function producing decorators."""
        super(GRPCClientUnaryOutputInterceptor, self).__init__()
        self._decorator_fn = decorator_fn

    def _intercept_call(self, continuation, client_call_details, request_or_iterator):
        """Interceptor implementation"""
        metadata = _get_metadata_map_from_client_details(client_call_details)
        decorator = self._decorator_fn(client_call_details.method, metadata)
        if not decorator:
            handler = continuation
        else:
            handler = decorator(continuation)

        return handler(client_call_details, request_or_iterator)

    def intercept_unary_unary(self, continuation, client_call_details, request):
        """Intercept unary-unary."""
        return self._intercept_call(continuation, client_call_details, request)

    def intercept_stream_unary(self, continuation, client_call_details, request_iterator):
        """Intercept stream-unary."""
        return self._intercept_call(continuation, client_call_details, request_iterator)


class GRPCClientMiddleware(object):
    """Base class for GRPC client-side middleware.

    GRPCMiddleware implementations must provide a get_decorator method:

    # def get_decorator(self, method_name, metadata)


    Which takes a string method name, and dict of rpc leading metadata and
    returns a decorator that can be applied to the underlying rpc method.
    Additionally:
      __init__ is guaranteed to be called before the server is started.
      get_interceptors(self) will be called to retrieve all GRPC interceptors
        necessary for the middleware.  Users may extend this method to include
        additional interceptors.
    """

    def __init__(self, client_label, server_label, interceptor_class):
        """Initialize"""
        super(GRPCClientMiddleware, self).__init__()
        self._server_label = server_label
        self._client_label = client_label
        self._interceptor_class = interceptor_class

    @property
    def server_label(self):
        """Get server label."""
        return self._server_label

    @property
    def client_label(self):
        """Get client label."""
        return self._client_label

    def get_interceptors(self):
        """Get a list of interceptors needed by the middleware."""
        return [self._interceptor_class(self.get_decorator)]


class ClientSideMetricsMiddleware(GRPCClientMiddleware):
    """GRPC middleware that captures prometheus metrics."""

    def __init__(self, client_label, server_label):
        """Initialize"""
        super(ClientSideMetricsMiddleware, self).__init__(
            client_label, server_label, GRPCClientGeneralInterceptor
        )

    class Timer(object):
        """Decorator that wraps a function in a prometheus histogram."""

        def __init__(self, histogram):
            """Initializes with the histogram object."""
            self._histogram = histogram

        def __call__(self, fn):
            """Wrap a method with a histogram."""

            @functools.wraps(fn)
            def wrap(request, context):
                """Inner wrapper."""
                with self._histogram.time():
                    return fn(request, context)

            return wrap

    def get_decorator(self, method_name, _):
        """Normalize metric name and return decorator that captures metrics."""
        service_label, endpoint_label = get_service_and_method_from_url(method_name)
        return self.Timer(
            CLIENTSIDE_METRICS_HISTO.labels(
                client_name=self.client_label,
                server_name=self.server_label,
                service=service_label,
                endpoint=endpoint_label,
            )
        )


class ClientSideExceptionCountMiddleware(GRPCClientMiddleware):
    """GRPC middleware that captures prometheus metrics for unary outputs."""

    def __init__(self, client_label, server_label):
        """Initialize"""
        super(ClientSideExceptionCountMiddleware, self).__init__(
            client_label, server_label, GRPCClientUnaryOutputInterceptor
        )

    class Counter(object):
        """Decorator that wraps a function in a exception counter."""

        def __init__(self, counter, client_name, server_name, service, endpoint):
            """Initializes with the counter object."""
            self._counter = counter
            self._client_name = client_name
            self._server_name = server_name
            self._service = service
            self._endpoint = endpoint

        def __call__(self, fn):
            """Wrap a method with an exception counter."""

            @functools.wraps(fn)
            def wrap(request, context):
                """Inner wrapper."""
                r = fn(request, context)
                if r.exception():
                    # If we get a Rendezvous error, we want some more information about the type
                    # of error we are getting. For example, a GRPC timeout error will be labelled as
                    # exception "_Rendezvous: <StatusCode.DEADLINE_EXCEEDED: 4>". All errors can be
                    # found at https://grpc.github.io/grpc/python/grpc.html#grpc-status-code
                    if type(r.exception()).__name__ == GRPC_RENDEZVOUS_ERROR:
                        exception = GRPC_RENDEZVOUS_ERROR + ": " + repr(r.exception().code())
                    # No guarantees of status code for other errors--only report error type.
                    else:
                        exception = type(r.exception()).__name__
                    self._counter.labels(
                        client_name=self._client_name,
                        server_name=self._server_name,
                        service=self._service,
                        endpoint=self._endpoint,
                        exception=exception,
                    ).inc()
                return r

            return wrap

    def get_decorator(self, method_name, _):
        """Normalize method name and return decorator that captures exceptions"""
        service_label, endpoint_label = get_service_and_method_from_url(method_name)
        return self.Counter(
            CLIENTSIDE_ERROR_COUNTER,
            self.client_label,
            self.server_label,
            service_label,
            endpoint_label,
        )


class ClientExceptionTranslationMiddlewareUnaryOutput(GRPCClientMiddleware):
    """Translate client exception"""

    def __init__(self, client_label, server_label, code_to_exception_class_func):
        """Initialize"""
        super(ClientExceptionTranslationMiddlewareUnaryOutput, self).__init__(
            client_label, server_label, GRPCClientUnaryOutputInterceptor
        )
        self._code_to_exception_class_func = code_to_exception_class_func

    class Translator(object):
        """Decorator that wraps a function in a exception translator"""

        def __init__(self, code_to_exception_class_func):
            """Initializes with the counter object"""
            self._code_to_exception_class_func = code_to_exception_class_func

        def __call__(self, fn):
            """Wrap a method with an exception counter"""

            @functools.wraps(fn)
            def wrap(request, context):
                """Execute a function, if an exception is raised, change its type if necessary"""
                try:
                    result = fn(request, context)
                    if result.code() is grpc.StatusCode.OK:
                        return result
                    else:
                        raise result
                except grpc.RpcError as exc:
                    raise_exception_from_grpc_exception(self._code_to_exception_class_func, exc)

            return wrap

    def get_decorator(self, method_name, _):
        """Return exception translator decorator"""
        return self.Translator(self._code_to_exception_class_func)


class ClientRetryingMiddlewareUnaryOutput(GRPCClientMiddleware):
    """Translate client exception"""

    def __init__(self, client_label, server_label, exceptions_to_retry, max_retries):
        """Initialize"""
        super(ClientRetryingMiddlewareUnaryOutput, self).__init__(
            client_label, server_label, GRPCClientUnaryOutputInterceptor
        )
        self._exceptions_to_retry = exceptions_to_retry
        self._max_retries = max_retries

    class Retrier(object):
        """Decorator that wraps a function in a exception translator"""

        def __init__(self, exceptions_to_retry, max_retries):
            """Initializes with the counter object"""
            self._exceptions_to_retry = exceptions_to_retry
            self._max_retries = max_retries

        def __call__(self, fn):
            """Wrap a method with an exception counter"""
            return backoff.on_exception(backoff.expo, self._exceptions_to_retry, self._max_retries)(
                fn
            )

    def get_decorator(self, method_name, _):
        """Return exception translator decorator"""
        return self.Retrier(self._exceptions_to_retry, self._max_retries)


def raise_exception_from_grpc_exception(code_to_exception_class_func, exc):
    """Raise exception from exc, translating with code_to_exception_class_func"""
    code = None
    details = "[]"  # Details are expected to be jsondeserializable

    if exc.code() == grpc.StatusCode.DEADLINE_EXCEEDED:
        raise TimeoutError()
    elif exc.code() == grpc.StatusCode.UNIMPLEMENTED:
        raise NotImplementedError()
    elif exc.code() == grpc.StatusCode.UNAVAILABLE:
        raise ConnectionRefusedError()

    for key, value in exc.trailing_metadata():
        if key == "error_code":
            try:
                code = int(value)
            except (TypeError, ValueError):
                pass
        elif key == "error_details":
            details = value

    if code_to_exception_class_func:
        exception_class = code_to_exception_class_func(code)

        if exception_class:
            exception_args = json.loads(details)
            raise exception_class(*exception_args)
    raise exc


def _get_metadata_map_from_client_details(client_call_details):
    """Get metadata key->value map from client_call_details"""
    metadata = {metadatum[0]: metadatum[1] for metadatum in (client_call_details.metadata or [])}
    return metadata



# Copyright 2020-present Kensho Technologies, LLC.
import unittest


class ImportTests(unittest.TestCase):
    def test_base_imports(self):
        """Test that the base file imports successfully."""
        from ...server import base  # noqa

    def test_middleware_imports(self):
        """Test that the middleware file imports successfully."""
        from ...server import middleware  # noqa



# Copyright 2020-present Kensho Technologies, LLC.
import unittest
from unittest.mock import call, create_autospec

from google.protobuf import descriptor_pb2
from google.protobuf.descriptor_pool import DescriptorPool
from grpc_reflection.v1alpha import reflection_pb2

from eagr.reflection import reflection_descriptor_database
from eagr.tests.reflection import utils


class TestReflectionDescriptorDatabase(unittest.TestCase):
    def test_get_proto_bytes_for_requests(self):
        # Test empty case
        file_requests_empty = []
        proto_bytes_empty = reflection_descriptor_database.get_proto_bytes_for_requests(
            utils.reflection_client_mock, file_requests_empty
        )
        self.assertEqual([], proto_bytes_empty)

        # Test non-empty cases
        file_requests = (
            reflection_pb2.ServerReflectionRequest(file_by_filename=request)
            for request in ["proto_a", "proto_c"]
        )
        proto_bytes = reflection_descriptor_database.get_proto_bytes_for_requests(
            utils.reflection_client_mock, file_requests
        )
        self.assertIn(utils.PROTO_BYTES_A, proto_bytes)
        self.assertIn(utils.PROTO_BYTES_C, proto_bytes)
        self.assertNotIn(utils.PROTO_BYTES_B, proto_bytes)

    def test_import_dependencies_then_proto(self):
        descriptor_pool_mock = create_autospec(spec=DescriptorPool(), spec_set=True)
        # Test detached proto_a
        parsed_proto_bytes_a = descriptor_pb2.FileDescriptorProto()
        parsed_proto_bytes_a.ParseFromString(utils.PROTO_BYTES_A)
        reflection_descriptor_database.import_dependencies_then_proto(
            utils.reflection_client_mock,
            descriptor_pool_mock,
            "proto_a",
            set(),
            {"proto_a": parsed_proto_bytes_a},
        )
        expected_calls = [call(utils.PROTO_A)]
        descriptor_pool_mock.Add.assert_has_calls(expected_calls, any_order=False)

        # Test proto_b including dependencies in a tree
        descriptor_pool_mock = create_autospec(spec=DescriptorPool(), spec_set=True)
        parsed_proto_bytes_b = descriptor_pb2.FileDescriptorProto()
        parsed_proto_bytes_b.ParseFromString(utils.PROTO_BYTES_B)
        reflection_descriptor_database.import_dependencies_then_proto(
            utils.reflection_client_mock,
            descriptor_pool_mock,
            "proto_b",
            set(),
            {"proto_b": parsed_proto_bytes_b},
        )

        expected_calls = [
            call(utils.PROTO_C),
            call(utils.PROTO_E),
            call(utils.PROTO_F),
            call(utils.PROTO_D),
            call(utils.PROTO_B),
        ]
        descriptor_pool_mock.Add.assert_has_calls(expected_calls, any_order=False)

    def test_add_protos_to_descriptor_pool(self):
        descriptor_pool_mock = create_autospec(spec=DescriptorPool(), spec_set=True)

        # Test empty case
        reflection_descriptor_database.add_protos_to_descriptor_pool(
            utils.reflection_client_mock, descriptor_pool_mock, []
        )
        self.assertEqual(len(descriptor_pool_mock.Add.call_args_list), 0)

        # Test non-empty case
        reflection_descriptor_database.add_protos_to_descriptor_pool(
            utils.reflection_client_mock,
            descriptor_pool_mock,
            [utils.PROTO_BYTES_A, utils.PROTO_BYTES_B],
        )
        expected_calls = [
            call(utils.PROTO_A),
            call(utils.PROTO_C),
            call(utils.PROTO_E),
            call(utils.PROTO_F),
            call(utils.PROTO_D),
            call(utils.PROTO_B),
        ]
        descriptor_pool_mock.Add.assert_has_calls(expected_calls, any_order=False)

    def test_build_database_from_stub(self):
        # Only need to test to make sure it runs without error
        reflection_descriptor_database.build_database_from_stub(utils.reflection_client_mock)



# Copyright 2020-present Kensho Technologies, LLC.
"""Helpers for testing grpc clients"""
from concurrent import futures
from contextlib import contextmanager

import grpc


CONNECTION_MASK = "localhost:{}"
GRPC_GRACE_PERIOD = 1  # seconds


@contextmanager
def inprocess_grpc_server(servicer, servicer_add_function, num_threads=1):
    """Create an in-process grpc server for the servicer for testing purposes

    This should not be used anywhere in production.

    Sample usage:

        class MyServicer(foo_pb2_grpc.ServiceServicer):
            def bar(self, req, context):
                return Bar()

        with inprocess_grpc_server(
            MyServicer(), foo_pb2_grpc.add_ServiceServicer_to_server
        ) as address:
            client = make_grpc_client("", "", address, foo_pb2_grpc.ServiceStub)
            bar = client.bar()

    Args:
        servicer: a grpc servicer class derived from the base generated by the compiler
        servicer_add_function: generated function for adding a servicer to the server
        num_threads: optional number of threads to run. 1 by default

    Yields:
        a connection string for the client to open an insecure channel
    """
    grpc_server = None
    try:
        thread_pool = futures.ThreadPoolExecutor(max_workers=num_threads)
        grpc_server = grpc.server(thread_pool)
        grpc_address = CONNECTION_MASK.format(0)
        servicer_add_function(servicer, grpc_server)
        actual_port = grpc_server.add_insecure_port(grpc_address)
        actual_connection_address = CONNECTION_MASK.format(actual_port)
        grpc_server.start()
        yield actual_connection_address
    finally:
        if grpc_server:
            event = grpc_server.stop(GRPC_GRACE_PERIOD)
            event.wait(GRPC_GRACE_PERIOD)



# Copyright 2020-present Kensho Technologies, LLC.
# @generated
# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: eagr/protos/test_service.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import wrappers_pb2 as google_dot_protobuf_dot_wrappers__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='test_service.proto',
  package='eagr',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=_b('\n\x12test_service.proto\x12\x04\x65\x61gr\x1a\x1egoogle/protobuf/wrappers.proto2\xc9\x02\n\x0bTestService\x12J\n\nUnaryUnary\x12\x1c.google.protobuf.StringValue\x1a\x1c.google.protobuf.StringValue\"\x00\x12M\n\x0bUnaryStream\x12\x1c.google.protobuf.StringValue\x1a\x1c.google.protobuf.StringValue\"\x00\x30\x01\x12M\n\x0bStreamUnary\x12\x1c.google.protobuf.StringValue\x1a\x1c.google.protobuf.StringValue\"\x00(\x01\x12P\n\x0cStreamStream\x12\x1c.google.protobuf.StringValue\x1a\x1c.google.protobuf.StringValue\"\x00(\x01\x30\x01\x62\x06proto3')
  ,
  dependencies=[google_dot_protobuf_dot_wrappers__pb2.DESCRIPTOR,])



_sym_db.RegisterFileDescriptor(DESCRIPTOR)



_TESTSERVICE = _descriptor.ServiceDescriptor(
  name='TestService',
  full_name='eagr.TestService',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  serialized_start=61,
  serialized_end=390,
  methods=[
  _descriptor.MethodDescriptor(
    name='UnaryUnary',
    full_name='eagr.TestService.UnaryUnary',
    index=0,
    containing_service=None,
    input_type=google_dot_protobuf_dot_wrappers__pb2._STRINGVALUE,
    output_type=google_dot_protobuf_dot_wrappers__pb2._STRINGVALUE,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='UnaryStream',
    full_name='eagr.TestService.UnaryStream',
    index=1,
    containing_service=None,
    input_type=google_dot_protobuf_dot_wrappers__pb2._STRINGVALUE,
    output_type=google_dot_protobuf_dot_wrappers__pb2._STRINGVALUE,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='StreamUnary',
    full_name='eagr.TestService.StreamUnary',
    index=2,
    containing_service=None,
    input_type=google_dot_protobuf_dot_wrappers__pb2._STRINGVALUE,
    output_type=google_dot_protobuf_dot_wrappers__pb2._STRINGVALUE,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='StreamStream',
    full_name='eagr.TestService.StreamStream',
    index=3,
    containing_service=None,
    input_type=google_dot_protobuf_dot_wrappers__pb2._STRINGVALUE,
    output_type=google_dot_protobuf_dot_wrappers__pb2._STRINGVALUE,
    serialized_options=None,
  ),
])
_sym_db.RegisterServiceDescriptor(_TESTSERVICE)

DESCRIPTOR.services_by_name['TestService'] = _TESTSERVICE

# @@protoc_insertion_point(module_scope)



# Copyright 2020-present Kensho Technologies, LLC.
import unittest

from google.protobuf.wrappers_pb2 import StringValue
import grpc

from ...client.client_test_helpers import inprocess_grpc_server
from ...protos import test_service_pb2_grpc


class Servicer(test_service_pb2_grpc.TestServiceServicer):
    """Stupid Servicer"""

    def UnaryUnary(self, req, context):
        """Reflection"""
        return req

    def UnaryStream(self, req, context):
        """Reflect 10 times"""
        for _ in range(10):
            yield req

    def StreamUnary(self, req, context):
        """Concatenation"""
        response = ""
        for x in req:
            response += x.value
        return StringValue(value=response)

    def StreamStream(self, req, context):
        """Reflection"""
        for x in req:
            yield x


class TestClientTestHelpers(unittest.TestCase):
    def test_methods(self):
        with inprocess_grpc_server(
            Servicer(), test_service_pb2_grpc.add_TestServiceServicer_to_server
        ) as address:
            channel = grpc.insecure_channel(address)
            stub = test_service_pb2_grpc.TestServiceStub(channel)

            unary_value = StringValue(value="foo")
            self.assertEqual(unary_value, stub.UnaryUnary(unary_value))
            self.assertEqual([unary_value] * 10, list(stub.UnaryStream(unary_value)))

            data_to_stream = [
                StringValue(value="foo"),
                StringValue(value="bar"),
                StringValue(value="baz"),
            ]

            self.assertEqual(
                StringValue(value="foobarbaz"), stub.StreamUnary((x for x in data_to_stream))
            )
            self.assertEqual(data_to_stream, list(stub.StreamStream((x for x in data_to_stream))))



# Copyright 2020-present Kensho Technologies, LLC.
"""Code for loading descriptor and symbol databases from grpc reflection service

This is Leonid's personal take on clean room implementation of
https://github.com/grpc/grpc/blob/master/test/cpp/util/proto_reflection_descriptor_database.cc
but with less bells and whistles.

In particular, the descriptor pool is filled in in a single shot, we do not hold client to
incrementally load additional data. Given that the only things we care about is the information
about all services on the server, we first ask for all the services, and then load data for each for
them using file_containing_symbol request.

The fear is that this might not be able to handle imports (i.e. what if the service depends on the
messages defined in other proto files). But this should not be a problem according to documentation
https://github.com/grpc/grpc/blob/master/doc/server-reflection.md in particular:
```
Because most usecases will require also requesting the transitive dependencies of requested files,
the queries will also return all transitive dependencies of the returned file.
Should interesting usecases for non-transitive queries turn up later, we can easily extend the
protocol to support them.
```
note that the response for file_containing_symbol request contains a list of descriptors, not a
single one.
"""

import funcy
from google.protobuf import descriptor_pb2
from google.protobuf.descriptor_pool import DescriptorPool
from google.protobuf.symbol_database import SymbolDatabase
from grpc_reflection.v1alpha import reflection_pb2, reflection_pb2_grpc


GET_SERVICES_REQUEST = reflection_pb2.ServerReflectionRequest(list_services="services")


def get_proto_bytes_for_requests(reflection_client, file_requests):
    """Return the file descriptor proto bytes list given file requests for services or files

    Args:
        reflection_client: ServerReflectionStub. GRPC reflection client
        file_requests: generator (ServerReflectionRequest). The requests for file descriptor
            protos to be queried on during server reflection

    Returns:
        bytes
    """
    file_descriptors_responses = reflection_client.ServerReflectionInfo(file_requests)
    proto_bytes = [
        file_descriptor_proto
        for response in file_descriptors_responses
        for file_descriptor_proto in response.file_descriptor_response.file_descriptor_proto
    ]
    return proto_bytes


def import_dependencies_then_proto(
    reflection_client, descriptor_pool, proto_name, imported_names, reflected_names
):
    """Recursively add file descriptor protos to descriptor pool, first importing dependencies

    Args:
        reflection_client: ServerReflectionStub. GRPC reflection client
        descriptor_pool: DescriptorPool. GRPC descriptor pool where file descriptor protos will be
            loaded
        proto_name: string. Name of the file desciptor proto to be imported
        imported_names: set. The names of the memoized, already imported file descriptor protos
        reflected_names: dict. Map of to-be-imported file descriptor proto names to their
            respective file descriptor protos

    Returns:
        None
    """
    if proto_name in imported_names:
        return

    # Anything in the proto_name depth-first search stack should already be reflected
    if proto_name not in reflected_names:
        raise RuntimeError(
            "Something went wrong. Stacked planned imports should either "
            "be already reflected or already imported. Please fix."
        )

    dependencies = reflected_names[proto_name].dependency  # pylint: disable=no-member
    # Reflect not-yet-reflected dependencies the same way we run server reflection on root protos
    not_yet_reflected_dependencies = []
    for dependency in dependencies:
        if (dependency not in imported_names) and (dependency not in reflected_names):
            not_yet_reflected_dependencies.append(dependency)

    if not_yet_reflected_dependencies:
        file_requests = (
            reflection_pb2.ServerReflectionRequest(file_by_filename=dependency)
            for dependency in not_yet_reflected_dependencies
        )
        file_descriptor_proto_bytes = get_proto_bytes_for_requests(reflection_client, file_requests)
        for serialized_proto in file_descriptor_proto_bytes:
            parsed_file_descriptor_proto = descriptor_pb2.FileDescriptorProto()
            parsed_file_descriptor_proto.ParseFromString(serialized_proto)
            file_descriptor_proto_name = (
                parsed_file_descriptor_proto.name  # pylint: disable=no-member
            )
            reflected_names[file_descriptor_proto_name] = parsed_file_descriptor_proto

    # Recursively import dependencies
    for dependency in dependencies:
        import_dependencies_then_proto(
            reflection_client, descriptor_pool, dependency, imported_names, reflected_names
        )

    # Import the proto itself, update memos
    descriptor_pool.Add(reflected_names[proto_name])
    imported_names.add(proto_name)
    del reflected_names[proto_name]


def add_protos_to_descriptor_pool(reflection_client, descriptor_pool, file_descriptor_proto_bytes):
    """Add protos and dependencies, bottom-up, to the protobuf descriptor pool

    Args:
        reflection_client: ServerReflectionStub. GRPC reflection client
        descriptor_pool: DescriptorPool. GRPC descriptor pool where file descriptor protos will be
            loaded
        file_descriptor_proto_bytes: list of proto_bytes strings for the root GRPC file descriptor
            proto files to be added to the descriptor pool

    Returns:
        None
    """
    # Set of already imported names
    imported_names = set()
    # Map of file descriptor proto names to the respective file descriptor protos
    reflected_names = {}
    # Stack of file descriptor proto names to be imported
    proto_names = []

    # Load the initial proto name stack and reflected names hash with root-level protos to import
    for serialized_proto in file_descriptor_proto_bytes:
        parsed_file_descriptor_proto = descriptor_pb2.FileDescriptorProto()
        parsed_file_descriptor_proto.ParseFromString(serialized_proto)
        file_descriptor_proto_name = parsed_file_descriptor_proto.name  # pylint: disable=no-member
        proto_names.append(file_descriptor_proto_name)
        reflected_names[file_descriptor_proto_name] = parsed_file_descriptor_proto

    # Recursively find file descriptor proto dependencies, and import the entire tree bottom-up
    for proto_name in proto_names:
        import_dependencies_then_proto(
            reflection_client, descriptor_pool, proto_name, imported_names, reflected_names
        )


def build_database_from_stub(reflection_client):
    """Build descriptor pool and symbol database from reflection service.

    Args:
        reflection_client: ServerReflectionStub. GRPC reflection client

    Returns:
        tuple (descriptor pool, symbol database)
    """
    services_response = funcy.first(
        # Note that this is stupid, but grpc has problems iterating over lists
        reflection_client.ServerReflectionInfo(x for x in [GET_SERVICES_REQUEST])
    )
    service_names = [service.name for service in services_response.list_services_response.service]
    file_requests = (
        reflection_pb2.ServerReflectionRequest(file_containing_symbol=service_name)
        for service_name in service_names
    )
    file_descriptor_proto_bytes = get_proto_bytes_for_requests(reflection_client, file_requests)

    descriptor_pool = DescriptorPool()
    add_protos_to_descriptor_pool(reflection_client, descriptor_pool, file_descriptor_proto_bytes)

    symbol_database = SymbolDatabase(descriptor_pool)
    return (descriptor_pool, symbol_database)


def build_database_from_channel(channel):
    """Build descriptor pool and symbol database from reflection service.

    Args:
        channel: GRPC channel

    Returns:
        tuple (descriptor pool, symbol database)
    """
    reflection_client = reflection_pb2_grpc.ServerReflectionStub(channel)
    return build_database_from_stub(reflection_client)



# Copyright 2020-present Kensho Technologies, LLC.
# Functionality to propagate exception information through the grpc
import json

import funcy
import grpc


def _is_grpc_unary_method(attr):
    """Check if attribute is a grpc method that returns unary."""
    return isinstance(attr, (grpc.UnaryUnaryMultiCallable, grpc.StreamUnaryMultiCallable))


def _is_grpc_stream_method(attr):
    """Check if attribute is a grpc method that returns a stream."""
    return isinstance(attr, (grpc.UnaryStreamMultiCallable, grpc.StreamStreamMultiCallable))


def _with_unary_function_error_translation(code_to_exception_class_func, func):
    """A decorator that translates server-side exceptions into clientside ones.

    Clientside jsonrpclib translates all unknown errors into ProtocolErrors initialized with
    a tuple (code, message). We assume that the errors have been encoded by flask_jsonrpc via
    with_server_error_translation and we can use code_to_exception_class_func to go from the code
    to an exception class that can be constructed with args deserialized from the message. If the
    exception cannot be retrieved, then we re-raise the protocol error.

    Args:
        code_to_exception_class_func: function that returns an exception corresponding to the code
                                      or None
        func: function to be returned

    Returns:
        wrapped function that extracts the server-side exceptions from those raised by jsonrpclib
    """

    @funcy.wraps(func)
    def decorated(*args, **kwargs):
        """Execute a function, if an exception is raised, change its type if necessary"""
        try:
            return func(*args, **kwargs)
        except grpc.RpcError as exc:
            raise_exception_from_grpc_exception(code_to_exception_class_func, exc)

    return decorated


def _with_generator_error_translation(code_to_exception_class_func, func):
    """Same wrapping as above, but for a generator"""

    @funcy.wraps(func)
    def decorated(*args, **kwargs):
        """Execute a function, if an exception is raised, change its type if necessary"""
        try:
            for x in func(*args, **kwargs):
                yield x
        except grpc.RpcError as exc:
            raise_exception_from_grpc_exception(code_to_exception_class_func, exc)

    return decorated


def raise_exception_from_grpc_exception(code_to_exception_class_func, exc):
    """Raise exception from exc, translating with code_to_exception_class_func"""
    code = None
    details = "[]"  # Details are expected to be jsondeserializable

    for key, value in exc.trailing_metadata():
        if key == "error_code":
            try:
                code = int(value)
            except (TypeError, ValueError):
                pass
        elif key == "error_details":
            details = value

    exception_class = code_to_exception_class_func(code)

    if exception_class:
        exception_args = json.loads(details)
        raise exception_class(*exception_args)
    else:
        raise exc


def update_grpc_stub_with_wrappers(stub, code_to_exception_class_func=None):
    """Update the stub's GRPC methods by applying decorators"""
    for attr_name in dir(stub):
        attr = getattr(stub, attr_name)

        if code_to_exception_class_func:
            if _is_grpc_unary_method(attr):
                setattr(
                    stub,
                    attr_name,
                    _with_unary_function_error_translation(code_to_exception_class_func, attr),
                )
            elif _is_grpc_stream_method(attr):
                setattr(
                    stub,
                    attr_name,
                    _with_generator_error_translation(code_to_exception_class_func, attr),
                )
            else:
                pass

    return stub



# Copyright 2020-present Kensho Technologies, LLC.
from collections import OrderedDict
import unittest

from google.protobuf import any_pb2, wrappers_pb2

from eagr.grpc_utils import response_translation


class TestResponseTranslation(unittest.TestCase):
    def test_translation_of_search_hit_with_primitives(self):
        """Verify that well-known types get translated into their values rather than into dicts."""

        values_and_serializations = (
            (wrappers_pb2.BoolValue(value=True), True),
            (wrappers_pb2.Int32Value(value=1), 1),
            (wrappers_pb2.UInt32Value(value=2), 2),
            # int64 is expected to be converted to a string because of JSON!!
            (wrappers_pb2.Int64Value(value=3), "3"),
            (wrappers_pb2.StringValue(value="abc"), "abc"),
        )

        for value, expected in values_and_serializations:
            any_proto = any_pb2.Any()
            any_proto.Pack(value)
            self.assertEqual(expected, response_translation.translate_message_to_dict(any_proto))

    def test_translation_of_datafixture(self):
        self.maxDiff = None
        raw_response = {
            "displayFieldDescriptors": [
                {
                    "description": "",
                    "fieldName": "profile_link",
                    "fieldType": "Hyperlink",
                    "name": "Company",
                    "subtextFieldName": "",
                    "subtextFieldType": "",
                },
                {
                    "description": "",
                    "fieldName": "geography",
                    "fieldType": "string",
                    "name": "Geography",
                    "subtextFieldName": "",
                    "subtextFieldType": "",
                },
            ],
            "searchHits": [
                {
                    "displayData": {
                        "geography": OrderedDict(
                            [
                                ("@type", "type.googleapis.com/google.protobuf.StringValue"),
                                ("value", "California, United States"),
                            ]
                        ),
                        "profile_link": OrderedDict(
                            [
                                ("@type", "type.googleapis.com/search_structs.Hyperlink"),
                                ("text", "Apple Inc. (XNAS:AAPL)"),
                                ("url", "url"),
                                ("highlights", []),
                            ]
                        ),
                    },
                    "hitId": "4004205",
                },
                {
                    "displayData": {
                        "geography": OrderedDict(
                            [
                                ("@type", "type.googleapis.com/google.protobuf.StringValue"),
                                ("value", "Virginia, United States"),
                            ]
                        ),
                        "profile_link": OrderedDict(
                            [
                                ("@type", "type.googleapis.com/search_structs.Hyperlink"),
                                ("text", "Apple Hospitality REIT, Inc. (XNYS:APLE)"),
                                ("url", "url"),
                                ("highlights", []),
                            ]
                        ),
                    },
                    "hitId": "4187996",
                },
            ],
            "totalResults": 2,
            "verticalName": "companies",
        }

        expected_response = {
            "displayFieldDescriptors": [
                {
                    "description": "",
                    "fieldName": "profile_link",
                    "fieldType": "Hyperlink",
                    "name": "Company",
                    "subtextFieldName": "",
                    "subtextFieldType": "",
                },
                {
                    "description": "",
                    "fieldName": "geography",
                    "fieldType": "string",
                    "name": "Geography",
                    "subtextFieldName": "",
                    "subtextFieldType": "",
                },
            ],
            "searchHits": [
                {
                    "displayData": {
                        "geography": "California, United States",
                        "profile_link": {
                            "highlights": [],
                            "text": "Apple Inc. " "(XNAS:AAPL)",
                            "url": "url",
                        },
                    },
                    "hitId": "4004205",
                },
                {
                    "displayData": {
                        "geography": "Virginia, United States",
                        "profile_link": {
                            "highlights": [],
                            "text": "Apple Hospitality " "REIT, Inc. " "(XNYS:APLE)",
                            "url": "url",
                        },
                    },
                    "hitId": "4187996",
                },
            ],
            "totalResults": 2,
            "verticalName": "companies",
        }

        translated_response = response_translation._get_clean_dict_value(raw_response)
        self.assertEqual(translated_response, expected_response)



# Copyright 2020-present Kensho Technologies, LLC.
import unittest

import grpc

from ...client.client_wrapper import (
    _with_generator_error_translation,
    _with_unary_function_error_translation,
    raise_exception_from_grpc_exception,
)


class FakeRpcError(grpc.RpcError):
    def __init__(self, code, details):
        """Create a fake rpc error with given code and details"""
        self._trailing_metadata = {}
        self._trailing_metadata["error_code"] = code
        self._trailing_metadata["error_details"] = details

    def trailing_metadata(self):
        """Return trailing metadata for the exception"""
        return self._trailing_metadata.items()


class ExceptionOne(Exception):
    """A dummy exception"""


def code_to_class(code):
    """Fake code to class"""
    if code == -1:
        return ExceptionOne

    return None


class ClientWrapperTests(unittest.TestCase):
    def test_raise_exception_from_grpc_exception(self):
        """Test that exceptions get converted properly"""

        # Verify that gRPC error that has no translation table, gets raised as a grpc exception
        non_registered_exc = FakeRpcError(-2, '["some details"]')  # Non-registered exception

        with self.assertRaises(grpc.RpcError):
            raise_exception_from_grpc_exception(code_to_class, non_registered_exc)

        exc = FakeRpcError(-1, '["some details"]')
        try:
            raise_exception_from_grpc_exception(code_to_class, exc)
        except ExceptionOne as e:
            self.assertIn("some details", str(e))

    def test_with_unary_function_error_translation(self):
        """Test that client errors get translated correctly"""

        def fail():
            """Raise an error invoking"""
            raise FakeRpcError(-1, '["some more details"]')

        decorated_fail = _with_unary_function_error_translation(code_to_class, fail)

        try:
            decorated_fail()
        except ExceptionOne as e:
            self.assertIn("some more details", str(e))

    def test_generator_wrapper(self):
        """Test wrapping a generator"""

        def fail_yielding():
            """Yield one item then raise an error"""
            yield 0
            raise FakeRpcError(-1, '["failed yielding"]')

        decorated_fail = _with_generator_error_translation(code_to_class, fail_yielding)
        gen = decorated_fail()

        self.assertEqual(next(gen), 0)

        try:
            next(gen)
        except ExceptionOne as e:
            self.assertIn("failed yielding", str(e))



# Copyright 2020-present Kensho Technologies, LLC.
from typing import Callable, Dict, Optional, Tuple

import backoff
from google.protobuf import json_format
import grpc

from eagr.grpc_utils import method, response_translation
from eagr.reflection import reflection_descriptor_database


MAX_RETRIES = 3


def _make_json_to_json_method_invocation(
    method_callable: Callable, proto_type: Callable, max_retries: int
) -> Callable:
    """Make function wrapping grpc method into json conversion"""
    # This invocation definition serves as a lambda for a GRPC method invocation
    @backoff.on_exception(backoff.expo, (grpc.RpcError,), max_retries)
    def invocation(
        input_dict: Dict,
        timeout: Optional[int] = None,
        metadata: Optional[Tuple] = None,
        credentials: Optional[grpc.CallCredentials] = None,
        wait_for_ready: Optional[bool] = None,
        compression: Optional[int] = None,
    ) -> Dict:
        """Convert input into protobuf, invoke grpc and convert response back to json"""
        input_message = proto_type()
        json_format.ParseDict(input_dict, input_message)
        response_as_message = method_callable(
            input_message, timeout=timeout, metadata=metadata, credentials=credentials
        )
        response_as_dict = response_translation.translate_message_to_dict(response_as_message)
        return response_as_dict

    return invocation


def make_json_grpc_client(host: str, service_name: str) -> Dict[str, Callable]:
    """Mount all GRPC methods for service onto client."""
    client = {}  # map method names to method invocations
    channel = grpc.insecure_channel(host, (("grpc.lb_policy_name", "round_robin"),))
    build_database_from_channel = reflection_descriptor_database.build_database_from_channel
    descriptor_pool_instance, symbol_database_instance = build_database_from_channel(channel)

    service_descriptor = descriptor_pool_instance.FindServiceByName(service_name)

    for method_descriptor in service_descriptor.methods:
        method_name = method_descriptor.name
        method_callable = method.make_grpc_unary_method(
            channel, service_name, method_descriptor, symbol_database_instance
        )
        input_type = method_descriptor.input_type
        input_prototype = symbol_database_instance.GetPrototype(input_type)
        method_invocation = _make_json_to_json_method_invocation(
            method_callable, input_prototype, max_retries=MAX_RETRIES
        )

        client[method_name] = method_invocation

    return client



# Copyright 2020-present Kensho Technologies, LLC.
from typing import Any, Dict

from google.protobuf import json_format


POD_TYPES = (int, float, str)
TYPE_KEY = "@type"
TYPE_KEY_SET = frozenset({TYPE_KEY})


def _get_clean_value(value):
    # type: (Any) -> Any
    """Determine which cleaner to use for the given value, and apply it."""
    if isinstance(value, dict):
        return _get_clean_dict_value(value)
    elif isinstance(value, list):
        return _get_clean_list_value(value)
    elif isinstance(value, POD_TYPES):  # Immutable built-in types.
        return value
    else:
        raise AssertionError(u"Cannot clean value of type {}: {}".format(type(value), value))


def _get_clean_dict_value(dct):
    # type: (Dict) -> Any
    """Clean up a dictionary which represents an encoded Any type.

    A dictionary can arise in 2 different ways when translating a message to a dictionary (via
    MessageToDict).

    The underlying data structure map be a dictionary (protobuf) or else the type that is being
    translated is an Any type. While translating an Any type the translator produces a dict
    that multiple keys one of which is a @type key that denotes the type of the message being
    translated.

    Examples:
        Any(Int32(value=3))

        Gets translated into a raw message of:

        {
            '@type': 'google...int32',
            'value': 3,
        }

        The desired output from this function is to replace the dictionary with the integer value 3.

        Any(SomeStructure(a=2, b=3))

        {
            '@type': 'google...some_structure',
            'a': 2,
            'b': 3
        }

        The desired output from this function is to retain the dictionary structure and instead
        simply remove the @type key. So

        {
            'a': 2,
            'b': 3
        }

    Args:
        dct: dict instance, as outputted from MessageToDict.

    Returns:
        dict, clean version of dict
    """
    keys = set(dct.keys())

    is_flat_any_type = TYPE_KEY in keys and len(keys) == 2

    if not is_flat_any_type:
        return _project_dict_without_type(dct)
    else:
        remaining_keys = set(keys) - TYPE_KEY_SET

        if len(remaining_keys) != 1:
            raise ValueError(
                u"Expected to find @type in the dict and one another key "
                u"Instead found {} for other keys.".format(remaining_keys)
            )

        remaining_key = list(remaining_keys)[0]
        value = dct[remaining_key]
        return _get_clean_value(value)


def _project_dict_without_type(dct):
    # type (Dict) -> Any
    """Clean a dict value.

    Args:
        dct: dict

    Returns:
        Any, any value to replace the dictionary with.
    """
    return {key: _get_clean_value(value) for key, value in dct.items() if key != TYPE_KEY}


def _get_clean_list_value(iterable):
    # type (List) -> Any
    """Clean a list iterable.

    Args:
        iterable: iterable

    Returns:
        Any, any value to replace the list with.
    """
    return [_get_clean_value(element) for element in iterable]


#
#  PUBLIC API
#


def translate_message_to_dict(message):
    # type (Any) -> Dict
    """Translate protobuf message to dictionary.

    Args:
        message: The protocol buffer message to be translated

    Returns:
        dict, clean version of the message in jsonifiably dict format.
    """
    # The default value is included here to make the response to the UI such that there are no
    # keys missing randomly. If this is changed, the UI wil have to start handling randomly
    # missing keys and the code that produces Any type conversion will convert Any types
    # inconsistently. Primitive Any types are flattened, while Any types that are structs
    # are kept as dicts currently with the logic relying on the number of keys.
    raw_response = json_format.MessageToDict(message, including_default_value_fields=True)
    simplified_dict = _get_clean_dict_value(raw_response)

    return simplified_dict



# Copyright 2020-present Kensho Technologies, LLC.
"""Testing various wrappers that make_grpc_client creates"""
import time
import unittest
from unittest.mock import MagicMock

from google.protobuf.wrappers_pb2 import StringValue
from opentracing import global_tracer, set_global_tracer
from prometheus_client.core import REGISTRY

from ...client import make_grpc_client
from ...client.client_test_helpers import inprocess_grpc_server
from ...protos import test_service_pb2_grpc


def _get_metric_value(metrics, metric_name, sample_name, labels):
    """Get value for sample of a metric"""
    for metric in metrics:
        if metric.name == metric_name:
            for sample in metric.samples:
                if sample.name == sample_name and sample.labels == labels:
                    return sample.value
    return None


class TestCallMetrics(unittest.TestCase):
    def test_call_metrics(self):
        servicer = MagicMock(test_service_pb2_grpc.TestServiceServicer)
        servicer.UnaryUnary = lambda x, _: x
        with inprocess_grpc_server(
            servicer, test_service_pb2_grpc.add_TestServiceServicer_to_server
        ) as address:
            set_global_tracer(global_tracer())
            client = make_grpc_client("foo", "bar", address, test_service_pb2_grpc.TestServiceStub)
            unary_value = StringValue(value="foo")
            self.assertEqual(unary_value, client.UnaryUnary(unary_value))
            labels = {
                "client_name": "foo",
                "server_name": "bar",
                "service": "eagr_TestService",
                "endpoint": "UnaryUnary",
            }
            call_count = REGISTRY.get_sample_value("clientside_grpc_endpoint_count", labels=labels)
            self.assertEqual(1, call_count)
            # Try that again to see if the number changes
            self.assertEqual(unary_value, client.UnaryUnary(unary_value))
            call_count = REGISTRY.get_sample_value("clientside_grpc_endpoint_count", labels=labels)
            self.assertEqual(2, call_count)

    def test_exception_tracking(self):
        servicer = MagicMock(test_service_pb2_grpc.TestServiceServicer)
        servicer.UnaryUnary = lambda x, _: time.sleep(10)
        count_labels = {
            "client_name": "foo",
            "server_name": "bar",
            "service": "eagr_TestService",
            "endpoint": "UnaryUnary",
        }
        exception_labels = dict(count_labels, exception="TimeoutError")

        with inprocess_grpc_server(
            servicer, test_service_pb2_grpc.add_TestServiceServicer_to_server
        ) as address:
            set_global_tracer(global_tracer())
            client = make_grpc_client("foo", "bar", address, test_service_pb2_grpc.TestServiceStub)
            calls_before = (
                REGISTRY.get_sample_value("clientside_grpc_endpoint_count", labels=count_labels)
                or 0
            )
            exceptions_before = (
                REGISTRY.get_sample_value(
                    "clientside_grpc_endpoint_error_total", labels=exception_labels
                )
                or 0
            )
            with self.assertRaises(TimeoutError):
                client.UnaryUnary(StringValue(value="foo"), timeout=1)
            calls_after = REGISTRY.get_sample_value(
                "clientside_grpc_endpoint_count", labels=count_labels
            )
            exceptions_after = REGISTRY.get_sample_value(
                "clientside_grpc_endpoint_error_total", labels=exception_labels
            )

            self.assertEqual(1, calls_after - calls_before)
            print(exceptions_after, exceptions_before)  # noqa
            self.assertEqual(1, exceptions_after - exceptions_before)



# Copyright 2020-present Kensho Technologies, LLC.
import cProfile
import functools
import json
import logging

from google.protobuf import json_format
from google.protobuf.message import Message as ProtoMessage
import grpc
from grpc import ServerInterceptor
import prometheus_client


logger = logging.getLogger(__name__)


GRPC_ENDPOINT_METRIC_NAME = "grpc_endpoint"
SERVICE_LABEL = "service"
ENDPOINT_LABEL = "endpoint"
ENDPOINT_METRIC_LABELS = (SERVICE_LABEL, ENDPOINT_LABEL)

METRICS_HISTO = prometheus_client.Histogram(
    GRPC_ENDPOINT_METRIC_NAME,
    "Response time histogram for grpc endpoints",
    labelnames=ENDPOINT_METRIC_LABELS,
)


def _wrap_rpc_handler(method_handler, wrapper):
    """Wrap a GRPC rpc handler object in a decorator

    GRPC rpc handlers carry around metadata in addition to the underlying
    method.  _wrap_rpc_handler takes care of the serialization metadata and
    underlying rpc type, just adding a decorator to the method.

    Args:
        method_handler: _InterceptorRpcMethodHandler
        wrapper: callable

    Returns:
        an rpc_method_handler
    """
    if method_handler.request_streaming:
        if method_handler.response_streaming:
            factory = grpc.stream_stream_rpc_method_handler
            fn = method_handler.stream_stream
        else:
            factory = grpc.stream_unary_rpc_method_handler
            fn = method_handler.stream_unary
    else:
        if method_handler.response_streaming:
            factory = grpc.unary_stream_rpc_method_handler
            fn = method_handler.unary_stream
        else:
            factory = grpc.unary_unary_rpc_method_handler
            fn = method_handler.unary_unary
    return factory(
        behavior=wrapper(fn),
        request_deserializer=method_handler.request_deserializer,
        response_serializer=method_handler.response_serializer,
    )


def _service_and_endpoint_labels_from_method(method_name):
    """Get normalized service_label, endpoint_label tuple from method name"""
    name_parts = method_name.split("/")
    if len(name_parts) != 3 or name_parts[0] != "" or name_parts[1] == "" or name_parts[2] == "":
        raise AssertionError("Invalid method name: {}".format(method_name))

    service_label = name_parts[1].replace(".", "_")
    endpoint_label = name_parts[2].replace(".", "_")
    return service_label, endpoint_label


class GRPCMiddleware(object):
    """Base class for GRPC middleware.

    GRPCMiddleware implementations must provide a get_decorator method:

    # def get_decorator(self, method_name, metadata)


    Which takes a string method name, and dict of rpc leading metadata and
    returns a decorator that can be applied to the underlying rpc method.
    Additionally:
      __init__ is guaranteed to be called before the server is started.
      get_interceptors(self) will be called to retrieve all GRPC interceptors
        necessary for the middleware.  Users may extend this method to include
        additional interceptors.
    """

    def get_interceptors(self):
        """Get a list of interceptors needed by the middleware."""
        return [self.MiddlewareInterceptor(self.get_decorator)]

    class MiddlewareInterceptor(ServerInterceptor):
        """Default GRPC interceptor used by middleware.  Applies a decorator"""

        def __init__(self, decorator_fn):
            """Initialize interceptor with a factory function producing decorators"""
            super(GRPCMiddleware.MiddlewareInterceptor, self).__init__()
            self._decorator_fn = decorator_fn

        def intercept_service(self, continuation, handler_call_details):
            """Interceptor implementation"""
            handler = continuation(handler_call_details)
            metadata = {
                metadatum.key: metadatum.value
                for metadatum in handler_call_details.invocation_metadata
            }
            decorator = self._decorator_fn(handler_call_details.method, metadata)
            # Note that handler may be None in which case we can't apply the
            # decorator and just propagate None
            if decorator and handler:
                handler = _wrap_rpc_handler(handler, decorator)
            return handler


class ProfilerMiddleware(GRPCMiddleware):
    """GRPC middleware that optionally profiles an RPC method"""

    class Profiler(object):
        """Profiling decorator"""

        def __init__(self, profile_mode):
            """Capture the profile mode (tottime or cumtime)"""
            self._profile_mode = profile_mode

        def __call__(self, fn):
            """Profile the rpc and print the resulting stats"""

            @functools.wraps(fn)
            def wrap(request, context):
                """Inner wrapper"""
                profile = cProfile.Profile()
                response = profile.runcall(fn, request, context)
                profile.print_stats(sort=self._profile_mode)
                return response

            return wrap

    def get_decorator(self, _, metadata):
        """If the client requests a profile return a decorator that profiles the RPC"""
        profile_mode = metadata.get("profile")
        if profile_mode and profile_mode in ("tottime", "cumtime"):
            logger.info("Profiling function invocation")
            return self.Profiler(profile_mode)
        elif profile_mode:
            logger.warning("Unknown profile mode {}. Skipping".format(profile_mode))
        return None


class MetricsMiddleware(GRPCMiddleware):
    """GRPC middleware that captures prometheus metrics"""

    def __init__(self):
        """Initialize"""
        super(MetricsMiddleware, self).__init__()

    class Timer(object):
        """Decorator that wraps a function in a prometheus histogram"""

        def __init__(self, histogram):
            """Initializes with the histogram object"""
            self._histogram = histogram

        def __call__(self, fn):
            """Wrap a method with a histogram"""

            @functools.wraps(fn)
            def wrap(request, context):
                """Inner wrapper"""
                with self._histogram.time():
                    return fn(request, context)

            return wrap

    def get_decorator(self, method_name, _):
        """Normalize metric name and return decorator that captures metrics"""
        # Make sure that the method name is valid
        service_label, endpoint_label = _service_and_endpoint_labels_from_method(method_name)
        return self.Timer(
            METRICS_HISTO.labels(**{SERVICE_LABEL: service_label, ENDPOINT_LABEL: endpoint_label})
        )


class ErrorMetaMiddleware(GRPCMiddleware):
    """GRPC middleware that translates exceptions into GRPC codes"""

    def __init__(self, exception_class_to_code_func):
        """Initialize middleware with a function that translates exceptions to codes"""
        self._exception_class_to_code_func = exception_class_to_code_func

    class ExceptionMapper(object):
        """Decorator that translates exceptions"""

        def __init__(self, mapper_func):
            """Capture mapper function"""
            self._mapper_func = mapper_func

        def __call__(self, fn):
            """Wrap a method with the exception translator"""

            @functools.wraps(fn)
            def wrap(request, context):
                """Inner wrapper"""
                try:
                    return fn(request, context)
                except Exception as e:
                    code = self._mapper_func(type(e))
                    details = json.dumps(list(e.args))
                    # If we have the code, use that. Otherwise we have to live with the default
                    if code:
                        context.set_trailing_metadata(
                            (("error_code", str(code)), ("error_details", details))
                        )
                        # Note that at this point GRPC will reset the code, but we can always hope
                        context.set_code(code)
                        context.set_details(details)
                    raise

            return wrap

    def get_decorator(self, _, __):
        """Return exception mapping decorator"""
        return self.ExceptionMapper(self._exception_class_to_code_func)


class LoggingMiddleware(GRPCMiddleware):
    """GRPC middleware that captures invocation logs."""

    def __init__(self, sanitizer=None):
        """Initialize"""
        super(LoggingMiddleware, self).__init__()
        self._sanitizer = sanitizer

    class Logger(object):
        """Decorator that wraps a function in a prometheus histogram"""

        def __init__(self, service, method, sanitizer):
            """Initializes with the service and method names"""
            self._service = service
            self._method = method
            self._sanitizer = sanitizer

        def __call__(self, fn):
            """Wrap a method with a histogram"""

            @functools.wraps(fn)
            def wrap(request, context):
                """Inner wrapper"""
                if isinstance(request, ProtoMessage):
                    sanitized_request = self._sanitizer(request) if self._sanitizer else request
                    logger.info(
                        "Invoked %s.%s(%s)",
                        self._service,
                        self._method,
                        str(json_format.MessageToDict(sanitized_request)),
                    )
                else:
                    logger.info(
                        "Invoked %s.%s with non-protobuf parameter", self._service, self._method
                    )

                return fn(request, context)

            return wrap

    def get_decorator(self, method_name, _):
        """Normalize metric name and return decorator that captures metrics"""
        # Make sure that the method name is valid
        service_label, endpoint_label = _service_and_endpoint_labels_from_method(method_name)
        return self.Logger(service_label, endpoint_label, self._sanitizer)



# Copyright 2020-present Kensho Technologies, LLC.
__version__ = "0.1"

from .client import make_grpc_client  # noqa
from .client.client_test_helpers import inprocess_grpc_server  # noqa
from .server import GRPCBase, run_grpc_servers  # noqa



# Copyright 2020-present Kensho Technologies, LLC.
"""Generic code that allows mounting grpc service calls on "REST" passthroughs"""

import flask
from google.protobuf import descriptor_pool, json_format, symbol_database

# Fetch a selected set of primitive types for which the protodict clean up involves
# a flattening of the namespace.
from eagr.flask_bridge.response_translation import translate_message_to_dict
from eagr.grpc_utils.method import make_grpc_unary_method
from eagr.reflection.reflection_descriptor_database import build_database_from_channel


def dict_to_grpc_method_handler(method, input_type, input_dict, symbol_database_instance):
    # type (Callable, MessageDescriptor, Dict[str, Any], Any) -> Dict
    """Invoke method identified by method_name of stub by converting dict payload into protobuf.

    Can handle only methods that accept a single protobuf as an input and produce a single
    protobuf (not stream) as an output

    Args:
        method: method to invoke
        input_type: message descriptor for the input proto
        input_dict: dictionary containing input data (that can be converted into proto using
            json_format.ParseDict)
        symbol_database_instance: symbol_database_instance

    Returns:
        dict containing the response
    """
    # Note that this is smart enough to process even well-known types e.g. StringValue
    input_prototype = symbol_database_instance.GetPrototype(input_type)
    input_message = input_prototype()
    json_format.ParseDict(input_dict, input_message)
    response_as_message = method(input_message)
    response_as_dict = translate_message_to_dict(response_as_message)
    return response_as_dict


def make_flask_dict_handler(method, func_name, input_type, symbol_database_instance):
    # type (Channel, str, MessageDescriptor, Any) -> Callable
    """Make a flask handler for the method.

    Args:
        method: method
        func_name: function name.
        input_type: type of input
        symbol_database_instance: symbol db instance
    """

    def method_handler():
        """Method handler."""
        input_dict = flask.request.get_json()
        response = dict_to_grpc_method_handler(
            method, input_type, input_dict, symbol_database_instance
        )
        return flask.json.jsonify(response)

    method_handler.__name__ = func_name
    return method_handler


def map_and_mount(
    flask_app,
    channel,
    service_name,
    json_service_path,
    descriptor_pool_instance=None,
    symbol_database_instance=None,
):
    # type (flask.Flask, Channel, str, str, Any, Any) -> None
    """Mount all json passthrough methods on specific path.

    Args:
        flask_app: flask app
        channel: grpc channel
        service_name: service name
        json_service_path: jsons service path
        descriptor_pool_instance: descriptor pool instance
        symbol_database_instance: symbol db instance
    """
    if descriptor_pool_instance is None:
        descriptor_pool_instance = descriptor_pool.Default()
    if symbol_database_instance is None:
        symbol_database_instance = symbol_database.Default()
    service_descriptor = descriptor_pool_instance.FindServiceByName(service_name)
    for method_descriptor in service_descriptor.methods:
        method_name = method_descriptor.name
        method_callable = make_grpc_unary_method(
            channel, service_name, method_descriptor, symbol_database_instance
        )
        input_type = method_descriptor.input_type
        method_handler = make_flask_dict_handler(
            method_callable, method_name, input_type, symbol_database_instance
        )
        method_route = "{}/{}".format(json_service_path, method_name)
        flask_app.route(method_route, methods=["POST"])(method_handler)


def map_and_mount_remote_server(flask_app, channel, service_name, json_service_path):
    # type (flask.Flask, Channel, str, str) -> None
    """Mount a remote service onto the flask app.

    Args:
        flask_app: flask_app
        channel: grpc channel
        service_name: service name
        json_service_path: json_service_path
    """
    descriptor_pool_instance, symbol_database_instance = build_database_from_channel(channel)
    return map_and_mount(
        flask_app,
        channel,
        service_name,
        json_service_path,
        descriptor_pool_instance=descriptor_pool_instance,
        symbol_database_instance=symbol_database_instance,
    )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['eagr',
 'eagr.client',
 'eagr.flask_bridge',
 'eagr.grpc_utils',
 'eagr.protos',
 'eagr.reflection',
 'eagr.server',
 'eagr.tests',
 'eagr.tests.client',
 'eagr.tests.grpc_utils',
 'eagr.tests.reflection',
 'eagr.tests.server']

package_data = \
{'': ['*']}

install_requires = \
['Flask>=0.12.2,<0.13.0',
 'backoff>=1.6.0,<2.0.0',
 'funcy>=1.7.2,<2.0.0',
 'grpcio-opentracing>=1.1,<2.0',
 'grpcio-reflection>=1.25,<2.0',
 'grpcio-tools>=1.25,<2.0',
 'grpcio>=1.25,<2.0',
 'opentracing>=2.2,<3.0',
 'opentracing_instrumentation>=3.2,<4.0',
 'prometheus_client>=0.7.1,<0.8.0',
 'protobuf>=3.6.0,<4.0.0',
 'pytz>=2019.3,<2020.0']

setup_kwargs = {
    'name': 'eagr',
    'version': '0.2.0',
    'description': 'A collection of utilities for making GRPC easier to use in python.',
    'long_description': None,
    'author': 'Louis Ades',
    'author_email': 'louis.ades@kensho.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



# Copyright 2020-present Kensho Technologies, LLC.
"""Code for generating gRPC stubs in Kensho-approved way"""
import itertools

import grpc

from eagr.client import client_side_middleware
from eagr.client.client_tracing import wrap_grpc_client_channel


DEFAULT_CHANNEL_OPTIONS = {
    # We want to round-robin between servers
    "grpc.lb_policy_name": "round_robin",
    # These options keep the connections alive and "warm" for the next request to succeed
    # when the client does not send a constant stream of requests.
    # This is done by periodically sending keepalive pings to the server that do two things:
    # (1) reming ELBs and whatever other systems are sitting between client and server do not
    # terminate "inactive" connections and
    # (2) restoring connection to another server replica if a server on the other end goes down
    # The flags are described here https://grpc.github.io/grpc/core/group__grpc__arg__keys.html
    # A good blog post for tuning is
    # https://cs.mcgill.ca/~mxia3/2019/02/23/Using-gRPC-in-Production/
    # Client should ping server to check for liveleness this often
    # This needs to happen more frequently than the ELB timeout
    "grpc.keepalive_time_ms": 120000,
    # Keep sending keepalive pings even if no requests are outstanding
    "grpc.keepalive_permit_without_calls": 1,
    # Try to keep the channel alive forever
    "grpc.http2.max_pings_without_data": 0,
    # Minimum time between sending pings (so that we don't send them too often)
    "grpc.http2.min_time_between_pings_ms": 120000,
}


def make_grpc_client(
    client_group,
    service_name,
    service_url,
    stub_cls,
    extra_channel_options=None,
    disable_tracing=False,
    code_to_exception_class_func=None,
    num_retries=3,
    exceptions_to_retry=None,
):
    """Generate a gRPC client with appropriate middleware and options.

    The client is described by the group, name and host:port string, where group and name are used
    for metrics and logging.

    Args:
        client_group: human readable description of the client group
        service_name: human-readable name of the service for metrics/logging purposes
        service_url: host:port string to connect to
        stub_cls: stub class to instantiate
        extra_channel_options: optional dict of grpc channel options as described in
        disable_tracing: boolean, set to disable tracing for this client
        code_to_exception_class_func: optional function for translating error codes to exceptions
        num_retries: number of times to retry (retriable) exceptions
        exceptions_to_retry: optional list of retriable exceptions. (ConnectionRefusedError,)
        by default

    Returns:
        an instance of the stub class
    """
    # Start by creating a set of options
    channel_options = dict(DEFAULT_CHANNEL_OPTIONS)

    if extra_channel_options:
        channel_options.update(extra_channel_options)

    channel = grpc.insecure_channel(service_url, options=tuple(channel_options.items()))

    # We retry connection refused errors (grpc.StatusCode.UNAVAILABLE) because those are
    # generally transient
    if exceptions_to_retry is None:
        exceptions_to_retry = (ConnectionRefusedError,)

    # Note that the middlewares are applied like decorators, so the later you are in the list
    # the earlier you are applied to the call
    middlewares = [
        client_side_middleware.ClientSideExceptionCountMiddleware(client_group, service_name),
        client_side_middleware.ClientRetryingMiddlewareUnaryOutput(
            client_group, service_name, exceptions_to_retry, num_retries
        ),
        client_side_middleware.ClientExceptionTranslationMiddlewareUnaryOutput(
            client_group, service_name, code_to_exception_class_func
        ),
        client_side_middleware.ClientSideMetricsMiddleware(client_group, service_name),
    ]
    interceptors = list(
        itertools.chain.from_iterable(middleware.get_interceptors() for middleware in middlewares)
    )
    decorated_channel = grpc.intercept_channel(channel, *interceptors)
    if disable_tracing:
        traced_channel = decorated_channel
    else:
        traced_channel = wrap_grpc_client_channel(decorated_channel)

    stub = stub_cls(traced_channel)

    # Keep the pointer to the channel inside the stub so that GC does not attempt to
    # collect it in the middle of interaction
    # cf. https://blog.jeffli.me/blog/2017/08/02/keep-python-grpc-client-connection-truly-alive/
    setattr(stub, "_channel_attribute_for_no_gc", decorated_channel)

    return stub



# Copyright 2020-present Kensho Technologies, LLC.
from grpc_opentracing import ActiveSpanSource, open_tracing_client_interceptor
from grpc_opentracing.grpcext import intercept_channel
import opentracing
from opentracing_instrumentation.request_context import get_current_span


class RequestContextSpanSource(ActiveSpanSource):
    """Implements interface of getting current span

    Gets span from RequestContext of opentracing instrumentation
    for the grpc opentracing library.
    """

    def get_active_span(self):
        """Get the request context active span"""
        return get_current_span()


def wrap_grpc_client_channel(channel):
    """Wraps a GRPC channel with tracing, given a global tracer has been registered"""
    if not opentracing.is_global_tracer_registered():
        raise Exception(
            "Global tracer has not been registered. Disable tracing or " "register a global tracer"
        )

    interceptor = open_tracing_client_interceptor(
        opentracing.global_tracer(), active_span_source=RequestContextSpanSource()
    )
    return intercept_channel(channel, interceptor)



# Copyright 2020-present Kensho Technologies, LLC.
from concurrent import futures
from contextlib import contextmanager
import logging
import warnings

import grpc
from grpc_reflection.v1alpha.reflection import enable_server_reflection
import prometheus_client


GRPC_REGISTRAR_ATTRIBUTE = "_REGISTRAR"
GRPC_TRACING_ATTRIBUTE = "_TRACING_ENABLED"
GRPC_GRACE_PERIOD = 10  # seconds

logger = logging.getLogger(__name__)


class GRPCBase(object):
    """Base class for simple GRPC servers.  Mostly a way to apply the metaclass"""

    _REGISTRAR = None  # Dont break the construction of this class with the assert

    def __init__(self):
        """Assert that the _REGISTRAR attribute exists"""
        super(GRPCBase, self).__init__()
        if not hasattr(self, GRPC_REGISTRAR_ATTRIBUTE):
            raise KeyError("GRPC implementation must define %s" % GRPC_REGISTRAR_ATTRIBUTE)


@contextmanager
def run_grpc_servers(
    servers,
    grpc_interface="0.0.0.0",
    grpc_port=7999,
    metrics_port=None,
    thread_pool=None,
    middlewares=None,
    grpc_server_options=None,
    enable_reflection_for_services=None,
    key_cert_pairs=None,
):
    """Run a bunch of GRPC servers

    Args:
        servers: Iterable of GRPCBase instances that will be exposed
        grpc_interface: Network interface to which grpc will be bound.  Probably shouldn't
                        be changed but if you only want to expose a service locally allows
                        for 127.0.0.1
        grpc_port: Port for GRPC requests (HTTP/2)
        metrics_port: Port for metrics (HTTP/1.1). Optional, must specify to enable metrics
        thread_pool: Thread pool for network requests.  Optional but the default is
                     unlikely to work for all use cases.
        middlewares: List of GRPCMiddleware objects
        grpc_server_options: an object that contains options directly passed to the grpc.server call
        enable_reflection_for_services: optional list of services for which to enable reflection
        key_cert_pairs: optional list of PEM encoded (key, cert_chain) pairs for TLS use
    """
    if thread_pool is None:
        warnings.warn("No thread pool specific - defaulting to 10 workers!")
        thread_pool = futures.ThreadPoolExecutor(max_workers=10)

    if middlewares is None:
        middlewares = []

    interceptors = []
    for middleware in middlewares:
        interceptors.extend(middleware.get_interceptors())

    grpc_server = grpc.server(thread_pool, interceptors=interceptors, options=grpc_server_options)

    try:
        for server in servers:
            getattr(server, GRPC_REGISTRAR_ATTRIBUTE)(grpc_server)
        grpc_address = grpc_interface + ":" + str(grpc_port)
        if key_cert_pairs:
            server_creds = grpc.ssl_server_credentials(key_cert_pairs)
            grpc_server.add_secure_port(grpc_address, server_creds)
        else:
            grpc_server.add_insecure_port(grpc_address)
        if enable_reflection_for_services is not None:
            enable_server_reflection(enable_reflection_for_services, grpc_server)
        grpc_server.start()
        if metrics_port is not None:
            prometheus_client.start_http_server(metrics_port)
        yield None
    finally:
        event = grpc_server.stop(GRPC_GRACE_PERIOD)
        event.wait(GRPC_GRACE_PERIOD)



# Copyright 2020-present Kensho Technologies, LLC.
import collections
import unittest
from unittest.mock import MagicMock, patch

from google.protobuf.util import json_format_proto3_pb2

from ...reflection import grpc_reflection_interface, reflection_descriptor_database
from ...tests.reflection import utils


# Mocks to patch out context-specific method calls in eagr.reflection.grpc_reflection_interface
def build_database_from_channel_mock(channel):
    """Returns a built DescriptorPool from a mock ServerReflectionStub, given any channel"""
    return reflection_descriptor_database.build_database_from_stub(utils.reflection_client_mock)


def find_service_by_name_mock(service_name):
    """Returns a mock GRPC service given a valid service name"""
    method_mock = MagicMock()
    method_mock.name = "MyTestMethod"
    method_mock.input_type = MagicMock()
    method_mock.input_type.prototype = build_database_from_channel_mock

    service_mock = MagicMock()
    service_mock.methods = [method_mock]
    if service_name == "good_service":
        return service_mock

    else:
        raise KeyError("Bad service {} given".format(service_name))


def make_grpc_unary_method_mock(channel, service_name, method_descriptor, symbol_database_instance):
    """Returns a mock callable method regardless of params provided"""

    def callable_mock():
        """Return a mock string value invariably"""
        return "return_value"

    return callable_mock


def get_prototype_mock(input_type):
    """Returns a mock callable method regardless of params provided"""

    def callable_mock():
        """Return a mock string value invariably"""
        return "return_value"

    return callable_mock


class ContextMock:
    def __init__(self, metadata):
        metadatum_tuple = collections.namedtuple("_Metadatum", ("key", "value"))
        self._metadata = map(metadatum_tuple._make, metadata)

    def invocation_metadata(self):
        return self._metadata


def method_to_callable_mock(grpc_method):
    def transformed_invocation(
        input_message,
        timeout=None,
        metadata=None,
        credentials=None,
        wait_for_ready=None,
        compression=None,
    ):
        return grpc_method(input_message, ContextMock(metadata))

    return transformed_invocation


class TestGRPCReflectionInterface(unittest.TestCase):
    def test__make_json_to_json_method_invocation(self):
        expected_input = 481516

        # Derived from google.protobuf's test cases
        test_input_dict = {"int32Value": expected_input}
        test_message = json_format_proto3_pb2.TestMessage

        def get_input_and_metadata(request, context):
            metadata_first = next(context.invocation_metadata())
            return test_message(
                string_value="int32_value: {}, metadata {}: {}".format(
                    request.int32_value, metadata_first.key, metadata_first.value
                )
            )

        json_method_invocation = grpc_reflection_interface._make_json_to_json_method_invocation(
            method_to_callable_mock(get_input_and_metadata), test_message, 3
        )

        self.assertEqual(
            json_method_invocation(test_input_dict, metadata=(("hi", "there"),))["stringValue"],
            "int32_value: 481516, metadata hi: there",
        )

    @patch("eagr.reflection.reflection_descriptor_database.SymbolDatabase", autospec=True)
    @patch("eagr.grpc_utils.method.make_grpc_unary_method", autospec=True)
    @patch("eagr.reflection.reflection_descriptor_database.DescriptorPool", autospec=True)
    @patch(
        "eagr.reflection.reflection_descriptor_database.build_database_from_channel", autospec=True
    )
    def test_make_json_grpc_client(
        self,
        build_database_mock_method,
        descriptor_pool_mock,
        make_unary_method_mock_method,
        symbol_database_mock,
    ):
        build_database_mock_method.side_effect = build_database_from_channel_mock

        descriptor_pool_mock_return_value = MagicMock()
        descriptor_pool_mock_return_value.FindServiceByName.side_effect = find_service_by_name_mock
        descriptor_pool_mock.return_value = descriptor_pool_mock_return_value

        make_unary_method_mock_method.side_effect = make_grpc_unary_method_mock

        symbol_database_mock_return_value = MagicMock()
        symbol_database_mock_return_value.GetPrototype.side_effect = get_prototype_mock
        symbol_database_mock.return_value.return_value = symbol_database_mock_return_value

        my_client = grpc_reflection_interface.make_json_grpc_client("my_host", "good_service")

        self.assertIn("MyTestMethod", my_client)

        with self.assertRaises(KeyError) as context:
            my_client = grpc_reflection_interface.make_json_grpc_client("my_host", "service_b")

        self.assertIn("service_b", context.exception.args[0])



# Copyright 2020-present Kensho Technologies, LLC.
def make_grpc_unary_method(channel, service_name, method_descriptor, symbol_database_instance):
    # type (Channel, str, MethodDescriptor, Any) -> Callable
    """Make grp callable on the channel.

    Args:
        channel: grpc channel
        service_name: name of service
        method_descriptor: method descriptor
        symbol_database_instance: symbol db instance
    """
    input_prototype = symbol_database_instance.GetPrototype(method_descriptor.input_type)
    output_prototype = symbol_database_instance.GetPrototype(method_descriptor.output_type)
    method = channel.unary_unary(
        "/{}/{}".format(service_name, method_descriptor.name),
        request_serializer=input_prototype.SerializeToString,
        response_deserializer=output_prototype.FromString,
    )
    return method
